"""

Sentence Splitter using PySBD

@shannons, @kylel

"""

import itertools
from typing import List, Tuple

import numpy as np
import pysbd

from papermage.magelib import (
    Document,
    Entity,
    PagesFieldName,
    Span,
    TokensFieldName,
    WordsFieldName,
)
from papermage.predictors import BasePredictor
from papermage.utils.merge import cluster_and_merge_neighbor_spans


class PysbdSentencePredictor(BasePredictor):
    """Sentence Boundary based on Pysbd

    Examples:
        >>> doc: Document = parser.parse("path/to/pdf")
        >>> predictor = PysbdSentenceBoundaryPredictor()
        >>> sentence_spans = predictor.predict(doc)
        >>> doc.annotate(sentences=sentence_spans)
    """

    @property
    def REQUIRED_DOCUMENT_FIELDS(self) -> List[str]:
        return [TokensFieldName]  # type: ignore

    def __init__(self) -> None:
        self._segmenter = pysbd.Segmenter(language="en", clean=False, char_span=True)

    def split_token_based_on_sentences_boundary(self, words: List[str]) -> List[Tuple[int, int]]:
        """
        Split a list of words into a list of (start, end) indices, indicating
        the start and end of each sentence.
        Duplicate of https://github.com/allenai/VILA/\blob/dd242d2fcbc5fdcf05013174acadb2dc896a28c3/src/vila/dataset/preprocessors/layout_indicator.py#L14      # noqa: E501

        Returns: List[Tuple(int, int)]
            a list of (start, end) for token indices within each sentence
        """

        if len(words) == 0:
            return [(0, 0)]
        combined_words = " ".join(words)

        char2token_mask = np.zeros(len(combined_words), dtype=np.int64)

        acc_word_len = 0
        for idx, word in enumerate(words):
            word_len = len(word) + 1
            char2token_mask[acc_word_len : acc_word_len + word_len] = idx
            acc_word_len += word_len

        segmented_sentences = self._segmenter.segment(combined_words)
        sent_boundary = [(ele.start, ele.end) for ele in segmented_sentences]

        split = []
        token_id_start = 0
        for start, end in sent_boundary:
            token_id_end = char2token_mask[start:end].max()
            if end + 1 >= len(char2token_mask) or char2token_mask[end + 1] != token_id_end:
                token_id_end += 1  # (Including the end)
            split.append((token_id_start, token_id_end))
            token_id_start = token_id_end
        return split

    def _predict(self, doc: Document) -> List[Entity]:
        if hasattr(doc, WordsFieldName):
            words = [word.text for word in getattr(doc, WordsFieldName)]
            attr_name = WordsFieldName
            # `words` is preferred as it should has better reading
            # orders and text representation
        else:
            words = [token.text for token in doc.tokens]
            attr_name = TokensFieldName

        split = self.split_token_based_on_sentences_boundary(words)

        sentence_spans: List[Entity] = []
        for start, end in split:
            if end - start == 0:
                continue
            if end - start < 0:
                raise ValueError

            cur_spans = getattr(doc, attr_name)[start:end]

            all_token_spans = list(itertools.chain.from_iterable([ele.spans for ele in cur_spans]))
            results = cluster_and_merge_neighbor_spans(all_token_spans)
            sentence_spans.append(Entity(spans=results.merged))

        return sentence_spans



"""

@kylel

"""


import unittest

from papermage.magelib import Span
from papermage.utils.merge import ClusterMergeResults, cluster_and_merge_neighbor_spans


class TestClusterMergeResults(unittest.TestCase):
    def test_num_clusters(self):
        result = ClusterMergeResults(
            items=["a", "b", "c"], cluster_ids=[[0, 1], [2]], cluster_id_to_merged={0: "ab", 1: "c"}
        )
        self.assertEqual(result.num_clusters, 2)

    def test_clusters(self):
        result = ClusterMergeResults(
            items=["a", "b", "c"], cluster_ids=[[0, 1], [2]], cluster_id_to_merged={0: "ab", 1: "c"}
        )
        self.assertListEqual(result.clusters, [["a", "b"], ["c"]])

    def test_merged(self):
        result = ClusterMergeResults(
            items=["a", "b", "c"], cluster_ids=[[0, 1], [2]], cluster_id_to_merged={0: "ab", 1: "c"}
        )
        self.assertListEqual(result.merged, ["ab", "c"])

    def test_get_merged(self):
        result = ClusterMergeResults(
            items=["a", "b", "c"], cluster_ids=[[0, 1], [2]], cluster_id_to_merged={0: "ab", 1: "c"}
        )
        self.assertEqual(result.get_merged("a"), "ab")
        self.assertEqual(result.get_merged("b"), "ab")
        self.assertEqual(result.get_merged("c"), "c")


class TestMergeClusterSpans(unittest.TestCase):
    def setUp(self):
        self.a = Span(0, 1)
        self.b = Span(1, 2)
        self.c = Span(3, 4)
        self.d = Span(4, 5)
        self.e = Span(9, 10)

    def test_empty(self):
        out = cluster_and_merge_neighbor_spans(spans=[])
        self.assertEqual(out.num_clusters, 0)
        self.assertListEqual(out.clusters, [])
        self.assertListEqual(out.merged, [])

    def test_single(self):
        out = cluster_and_merge_neighbor_spans(spans=[self.a])
        self.assertEqual(out.num_clusters, 1)
        self.assertListEqual(out.clusters, [[self.a]])
        self.assertListEqual(out.merged, [self.a])

    def test_disjoint(self):
        out = cluster_and_merge_neighbor_spans(spans=[self.a, self.c, self.e])
        self.assertEqual(out.num_clusters, 3)
        self.assertListEqual(out.clusters, [[self.a], [self.c], [self.e]])
        self.assertListEqual(out.merged, [self.a, self.c, self.e])
        # reversed order
        out = cluster_and_merge_neighbor_spans(spans=[self.e, self.c, self.a])
        self.assertEqual(out.num_clusters, 3)
        self.assertListEqual(out.clusters, [[self.a], [self.c], [self.e]])
        self.assertListEqual(out.merged, [self.a, self.c, self.e])

    def test_cluster(self):
        out = cluster_and_merge_neighbor_spans(spans=[self.a, self.b, self.c, self.d, self.e], distance=0)
        self.assertEqual(out.num_clusters, 3)
        self.assertListEqual(out.clusters, [[self.a, self.b], [self.c, self.d], [self.e]])
        self.assertListEqual(out.merged, [Span(0, 2), Span(3, 5), Span(9, 10)])
        out = cluster_and_merge_neighbor_spans(spans=[self.a, self.b, self.c, self.d, self.e], distance=1)
        self.assertEqual(out.num_clusters, 2)
        self.assertListEqual(out.clusters, [[self.a, self.b, self.c, self.d], [self.e]])
        self.assertListEqual(out.merged, [Span(0, 5), Span(9, 10)])
        out = cluster_and_merge_neighbor_spans(spans=[self.a, self.b, self.c, self.d, self.e], distance=4)
        self.assertEqual(out.num_clusters, 1)
        self.assertListEqual(out.clusters, [[self.a, self.b, self.c, self.d, self.e]])
        self.assertListEqual(out.merged, [Span(0, 10)])



"""


@chrisw, @kylel

"""


from abc import abstractmethod
from typing import List

import numpy as np
from ncls import NCLS

from .box import Box
from .entity import Entity


class Indexer:
    """Stores an index for a particular collection of Entities.
    Indexes in this library focus on *INTERSECT* relations."""

    @abstractmethod
    def find(self, query: Entity) -> List[Entity]:
        """Returns all matching Entities given a suitable query"""
        raise NotImplementedError()


class EntitySpanIndexer(Indexer):
    """
    Manages a data structure for locating overlapping Entity Spans.
    Builds a static nested containment list from Entity Spans
    and accepts other Entity as search probes.

    See: https://github.com/biocore-ntnu/ncls

    [citation]
    Alexander V. Alekseyenko, Christopher J. Lee;
    Nested Containment List (NCList): a new algorithm for accelerating interval query of genome
      alignment and interval databases, Bioinformatics,
    Volume 23, Issue 11, 1 June 2007, Pages 1386–1393, https://doi.org/10.1093/bioinformatics/btl647
    """

    def __init__(self, entities: List[Entity]) -> None:
        starts = []
        ends = []
        ids = []

        for id, entity in enumerate(entities):
            for span in entity.spans:
                starts.append(span.start)
                ends.append(span.end)
                ids.append(id)

        self._entities = entities
        self._index = NCLS(
            np.array(starts, dtype=np.int32), np.array(ends, dtype=np.int32), np.array(ids, dtype=np.int32)
        )

        self._ensure_disjoint()

    def _ensure_disjoint(self) -> None:
        """
        Constituent span groups must be fully disjoint.
        Ensure the integrity of the built index.
        """
        for entity in self._entities:
            for span in entity.spans:
                matches = [match for match in self._index.find_overlap(span.start, span.end)]
                match_ids = [
                    matched_id for _start, _end, matched_id in self._index.find_overlap(span.start, span.end)
                ]
                if len(match_ids) > 1:
                    matches = [self._entities[match_id].to_json() for match_id in match_ids]
                    raise ValueError(
                        f"Detected overlap! While processing the Span {span} as part of query Entity {entity.to_json()}, we found that it overlaps with existing Entity(s):\n"
                        + "\n".join([f"\t{i}\t{m} " for i, m in zip(match_ids, matches)])
                    )

    def find(self, query: Entity) -> List[Entity]:
        if not isinstance(query, Entity):
            raise TypeError(f"EntityIndexer only works with `query` that is Entity type")

        if not query.spans:
            return []

        matched_ids = set()

        for span in query.spans:
            for _start, _end, matched_id in self._index.find_overlap(span.start, span.end):
                matched_ids.add(matched_id)

        matched_entities = [self._entities[matched_id] for matched_id in matched_ids]

        # TODO: kylel - is this necessary? seems like already does this (see tests)
        # Retrieval above doesn't preserve document order; sort here
        # TODO: provide option to return matched span groups in same order as self._entities
        #   (the span groups the index was built with originally)
        return sorted(list(matched_entities))


class EntityBoxIndexer(Indexer):
    """
    Manages a data structure for locating overlapping BoxGroups.
    Builds a static nested containment list from BoxGroups
    and accepts other BoxGroups as search probes.

    @kylel
    """

    def __init__(self, entities: List[Entity], allow_overlap: bool = True) -> None:
        self._entities = entities

        self._box_id_to_entity_id = {}
        self._boxes = []
        box_id = 0
        for i, e in enumerate(entities):
            for box in e.boxes:
                self._boxes.append(box)
                self._box_id_to_entity_id[box_id] = i
                box_id += 1

        self._np_boxes_x1 = np.array([b.l for b in self._boxes])
        self._np_boxes_y1 = np.array([b.t for b in self._boxes])
        self._np_boxes_x2 = np.array([b.l + b.w for b in self._boxes])
        self._np_boxes_y2 = np.array([b.t + b.h for b in self._boxes])
        self._np_boxes_page = np.array([b.page for b in self._boxes])

        if not allow_overlap:
            self._ensure_disjoint()

    def _find_overlap_boxes(self, query: Box) -> List[int]:
        x1, y1, x2, y2 = query.xy_coordinates
        mask = (
            (self._np_boxes_x1 <= x2)
            & (self._np_boxes_x2 >= x1)
            & (self._np_boxes_y1 <= y2)
            & (self._np_boxes_y2 >= y1)
            & (self._np_boxes_page == query.page)
        )
        return np.where(mask)[0].tolist()

    def _find_overlap_entities(self, query: Box) -> List[int]:
        return [self._box_id_to_entity_id[box_id] for box_id in self._find_overlap_boxes(query)]

    def _ensure_disjoint(self) -> None:
        """
        Constituent box groups must be fully disjoint.
        Ensure the integrity of the built index.
        """
        for entity in self._entities:
            for box in entity.boxes:
                match_ids = self._find_overlap_entities(query=box)
                if len(match_ids) > 1:
                    matches = [self._entities[match_id].to_json() for match_id in match_ids]
                    raise ValueError(
                        f"Detected overlap! While processing the Box {box} as part of query Entity {entity.to_json()}, we found that it overlaps with existing Entity(s):\n"
                        + "\n".join([f"\t{i}\t{m} " for i, m in zip(match_ids, matches)])
                    )

    def find(self, query: Entity) -> List[Entity]:
        if not isinstance(query, Entity):
            raise TypeError(f"EntityBoxIndexer only works with `query` that is Entity type")

        if not query.boxes:
            return []

        match_ids = []
        for box in query.boxes:
            match_ids.extend(self._find_overlap_entities(query=box))

        return [self._entities[match_id] for match_id in sorted(set(match_ids))]



"""

Utilty methods related to Merging Entities. 

@kylel, @lucas

"""

from collections import defaultdict
from typing import Any, Dict, List, Tuple

from papermage.magelib import Span


class ClusterMergeResults:
    def __init__(
        self, items: List[Any], cluster_ids: List[List[int]], cluster_id_to_merged: Dict[int, Any]
    ) -> None:
        self.cluster_ids = cluster_ids
        self.items = items
        self.cluster_id_to_merged = cluster_id_to_merged
        self.item_id_to_cluster_id = {
            item_id: cluster_id for cluster_id, cluster in enumerate(cluster_ids) for item_id in cluster
        }

    def __repr__(self) -> str:
        return f"ClusterMergeResult(clusters={self.clusters}, cluster_id_to_merged={self.cluster_id_to_merged})"

    @property
    def num_clusters(self) -> int:
        return len(self.cluster_ids)

    @property
    def clusters(self) -> List[List[Any]]:
        """Instead of returning the indices of the items, return the items themselves."""
        if self.num_clusters:
            return [[self.items[item_id] for item_id in cluster] for cluster in self.cluster_ids]
        else:
            return []

    @property
    def merged(self) -> List[Any]:
        """Return the merged items."""
        if self.num_clusters:
            return [self.cluster_id_to_merged[cluster_id] for cluster_id in range(self.num_clusters)]
        else:
            return []

    def get_merged(self, item: Any) -> Any:
        """Return the merged item of the input item."""
        try:
            this_item_id = self.items.index(item)
        except ValueError:
            raise ValueError(f"Item {item} not found in the input items.")
        return self.cluster_id_to_merged[self.item_id_to_cluster_id[this_item_id]]


def cluster_and_merge_neighbor_spans(spans: List[Span], distance=1) -> ClusterMergeResults:
    """Merge neighboring spans in a list of un-overlapped spans:
    when the gaps between neighboring spans is smaller or equal to the
    specified distance, they are considered as the neighbors.

    Args:
        spans (List[Span]): The input list of spans.
        distance (int, optional):
            The upper bound of interval gaps between two neighboring spans.
            Defaults to 1.

    Returns:
        List[List[int]]: A list of clusters of neighboring spans. Each cluster is a list of indices of the spans.
        Dict[int, Span]: A dictionary mapping the cluster id to the merged span.
    """

    if len(spans) == 0:
        return ClusterMergeResults(items=[], cluster_ids=[], cluster_id_to_merged={})

    if len(spans) == 1:
        return ClusterMergeResults(items=spans, cluster_ids=[[0]], cluster_id_to_merged={0: spans[0]})

    # When sorted, only one iteration round is needed.
    spans = sorted(spans)

    is_neighboring_spans = (
        lambda span1, span2: min(abs(span1.start - span2.end), abs(span1.end - span2.start)) <= distance
    )

    # cluster and merge neighbors
    cur_cluster = 0
    cluster_ids = [[0]]
    cluster_id_to_big_span = {0: spans[0]}
    for span_id in range(1, len(spans)):
        cur_span = spans[span_id]
        cur_big_span = cluster_id_to_big_span[cur_cluster]
        if is_neighboring_spans(cur_span, cur_big_span):
            cluster_ids[cur_cluster].append(span_id)
            new_big_span = Span.create_enclosing_span(spans=[cur_big_span, cur_span])
            cluster_id_to_big_span[cur_cluster] = new_big_span
        else:
            cur_cluster += 1
            cluster_ids.append([span_id])
            cluster_id_to_big_span[cur_cluster] = cur_span

    return ClusterMergeResults(items=spans, cluster_ids=cluster_ids, cluster_id_to_merged=cluster_id_to_big_span)



"""

@kylel

"""

import json
import os
import pathlib

from papermage.magelib import Document
from papermage.recipes import CoreRecipe
from papermage.visualizers.visualizer import plot_entities_on_page

# load doc
recipe = CoreRecipe()
# pdfpath = pathlib.Path(__file__).parent.parent / "tests/fixtures/1903.10676.pdf"
# pdfpath = pathlib.Path(__file__).parent.parent / "tests/fixtures/2304.02623v1.pdf"
pdfpath = pathlib.Path(__file__).parent.parent / "tests/fixtures/2305.14772.pdf"
doc = recipe.from_pdf(pdf=pdfpath)

# visualize tokens
page_id = 0
plot_entities_on_page(page_image=doc.images[page_id], entities=doc.pages[page_id].tokens)

# visualize tables
page_id = 5
tables = doc.pages[page_id].intersect_by_box("tables")
plot_entities_on_page(page_image=doc.images[page_id], entities=tables)
for table in tables:
    print(table.text)

# visualize figures
figures = doc.pages[page_id].intersect_by_box("figures")
for figure in figures:
    print(figure.text)
plot_entities_on_page(page_image=doc.images[page_id], entities=figures)

# visualize blocks
blocks = doc.pages[page_id].intersect_by_box("blocks")
for block in blocks:
    print(block.text)
plot_entities_on_page(page_image=doc.images[page_id], entities=blocks)

# visualize sections
page_id = 4
sections = doc.pages[page_id].intersect_by_box("sections")
for section in sections:
    print(section.text)
plot_entities_on_page(page_image=doc.images[page_id], entities=sections)



import os
from glob import glob

from tqdm import tqdm

from papermage.rasterizers import PDF2ImageRasterizer

pdf_paths = [
    "tests/fixtures/2302.07302v1.pdf",
    "tests/fixtures/2020.acl-main.447.pdf",
    "tests/fixtures/2023.eacl-main.121.pdf",
    "tests/fixtures/2303.14334v2.pdf",
    "tests/fixtures/2304.02623v1.pdf",
    "tests/fixtures/papermage.pdf",
]
thumbnail_dir = "temp/"
os.makedirs(thumbnail_dir, exist_ok=True)

rasterizer = PDF2ImageRasterizer()

for pdf_path in tqdm(pdf_paths):
    pngfile = os.path.join(thumbnail_dir, os.path.basename(pdf_path).replace(".pdf", ".png"))
    if os.path.exists(pngfile):
        continue

    pdfimages = rasterizer.rasterize(input_pdf_path=pdf_path, dpi=200)
    first_page_img = pdfimages[0]

    first_page_img.save(pngfile, format="png")



"""

@kylel

"""

import pathlib
import unittest

from papermage.recipes import CoreRecipe


class TestCoreRecipe(unittest.TestCase):
    def setUp(self):
        self.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"
        self.recipe = CoreRecipe(
            svm_word_predictor_path=str(self.fixture_path / "svm_word_predictor/svm_word_predictor.tar.gz")
        )

    def test_stability(self):
        self.recipe.run(self.fixture_path / "1903.10676.pdf")
        # beyond summarization
        self.recipe.run(self.fixture_path / "2304.02623v1.pdf")
        # semantic reader
        # self.recipe.run(self.fixture_path / "2303.14334v2.pdf")
        # papermage
        # self.recipe.run(self.fixture_path / "papermage.pdf")
        # s2orc
        self.recipe.run(self.fixture_path / "2020.acl-main.447.pdf")
        self.recipe.run(self.fixture_path / "4be952924cd565488b4a239dc6549095029ee578.pdf")
        # longeval
        self.recipe.run(self.fixture_path / "2023.eacl-main.121.pdf")
        # citesee
        self.recipe.run(self.fixture_path / "2302.07302v1.pdf")



"""

An illustrataive example of how predictors can help each other.

@kylel

"""

import json
import os
import pathlib

from papermage.magelib import Box, Document, Entity, Metadata, Span
from papermage.recipes import CoreRecipe
from papermage.visualizers.visualizer import plot_entities_on_page


def fix_layout_preds_with_tokens(doc, page_id=1):
    # layoutparser doesnt produce very tight boundaries
    before = plot_entities_on_page(
        page_image=doc.images[page_id], entities=[block for block in doc.layout if block.boxes[0].page == page_id]
    )

    # let's use tokens to constrain the layout
    blocks = []
    for block in doc.layout:
        block.tokens = doc.find_by_box(query=Entity(boxes=[block.boxes[0]]), field_name="tokens")
        token_boxes = [box for t in block.tokens for box in t.boxes]
        new_block_box = Box.create_enclosing_box(boxes=token_boxes)
        new_block = Entity(boxes=[new_block_box], metadata=block.metadata)
        blocks.append(new_block)

    # fixed!
    after = plot_entities_on_page(
        page_image=doc.images[page_id], entities=[block for block in blocks if block.boxes[0].page == page_id]
    )
    return before, after


# get doc
recipe = CoreRecipe()
fixture_path = pathlib.Path(__file__).parent / "tests/fixtures"
doc = recipe.from_pdf(pdf=fixture_path / "2304.02623v1.pdf")

# analyze
before, after = fix_layout_preds_with_tokens(doc=doc, page_id=1)



"""

@kylel

"""

import unittest

from papermage.magelib import Box, Entity, Layer, Span


class TestLayer(unittest.TestCase):
    def setUp(self):
        e1 = Entity(spans=[Span(0, 1), Span(2, 3)])
        e2 = Entity(boxes=[Box(0, 0, 0, 0, 0), Box(1, 1, 1, 1, 1)])
        self.layer = Layer(entities=[e1, e2])

    def test_create(self):
        empty = Layer(entities=[])

    def test_index(self):
        self.assertEqual(self.layer[0], self.layer.entities[0])
        self.assertEqual(self.layer[1], self.layer.entities[1])

    def test_len(self):
        self.assertEqual(len(self.layer), 2)

    def test_iter(self):
        self.assertEqual(list(self.layer), self.layer.entities)

    def test_contains(self):
        self.assertIn(self.layer[0], self.layer)
        self.assertIn(self.layer[1], self.layer)

    def test_to_from_json(self):
        self.assertEqual(self.layer.to_json(), [e.to_json() for e in self.layer.entities])

        layer2 = Layer.from_json(self.layer.to_json())
        self.assertDictEqual(layer2[0].to_json(), self.layer[0].to_json())
        self.assertDictEqual(layer2[1].to_json(), self.layer[1].to_json())



"""

Block predictors.

@kylel

"""

import pathlib
import unittest

from papermage.predictors import LPEffDetPubLayNetBlockPredictor


class TestLayoutParserBoxPubLayNetBlockPredictor(unittest.TestCase):
    def setUp(self):
        self.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"
        self.predictor = LPEffDetPubLayNetBlockPredictor.from_pretrained()

    def test_predict(self):
        pass



"""

Predict visual blocks in a page.

@kylel

"""

from typing import Dict, Optional

from papermage.predictors.base_predictors.lp_predictors import LPPredictor


class LPEffDetPubLayNetBlockPredictor(LPPredictor):
    @classmethod
    def from_pretrained(
        cls,
        device: Optional[str] = None,
    ):
        return super().from_pretrained(
            config_path="lp://efficientdet/PubLayNet",
            device=device,
        )



"""

@kylel

"""

# document field names
SymbolsFieldName = "symbols"
ImagesFieldName = "images"
MetadataFieldName = "metadata"
EntitiesFieldName = "entities"
RelationsFieldName = "relations"

PagesFieldName = "pages"
TokensFieldName = "tokens"
RowsFieldName = "rows"
BlocksFieldName = "blocks"
WordsFieldName = "words"
SentencesFieldName = "sentences"
ParagraphsFieldName = "paragraphs"

# these come from vila
TitlesFieldName = "titles"
AuthorsFieldName = "authors"
AbstractsFieldName = "abstracts"
KeywordsFieldName = "keywords"
SectionsFieldName = "sections"
ListsFieldName = "lists"
BibliographiesFieldName = "bibliographies"
EquationsFieldName = "equations"
AlgorithmsFieldName = "algorithms"
FiguresFieldName = "figures"
TablesFieldName = "tables"
CaptionsFieldName = "captions"
HeadersFieldName = "headers"
FootersFieldName = "footers"
FootnotesFieldName = "footnotes"



"""
@kylel
"""


import os
from abc import abstractmethod
from pathlib import Path
from typing import Any

from papermage.magelib import Document


class Recipe:
    @abstractmethod
    def run(self, input: Any) -> Document:
        if isinstance(input, Path):
            if input.suffix == ".pdf":
                return self.from_pdf(pdf=input)
            if input.suffix == ".json":
                return self.from_json(doc=input)

            raise NotImplementedError("Filetype not yet supported.")

        if isinstance(input, Document):
            return self.from_doc(doc=input)

        if isinstance(input, str):
            if os.path.exists(input):
                input = Path(input)
                return self.run(input=input)
            else:
                return self.from_str(text=input)

        raise NotImplementedError("Document input not yet supported.")

    @abstractmethod
    def from_str(self, text: str) -> Document:
        raise NotImplementedError

    @abstractmethod
    def from_pdf(self, pdf: Path) -> Document:
        raise NotImplementedError

    @abstractmethod
    def from_json(self, json: str) -> Document:
        raise NotImplementedError

    @abstractmethod
    def from_doc(self, doc: Document) -> Document:
        raise NotImplementedError



from papermage.predictors.base_predictors.base_predictor import BasePredictor
from papermage.predictors.base_predictors.hf_predictors import HFBIOTaggerPredictor
from papermage.predictors.block_predictors import LPEffDetPubLayNetBlockPredictor
from papermage.predictors.formula_predictors import LPEffDetFormulaPredictor
from papermage.predictors.sentence_predictors import PysbdSentencePredictor
from papermage.predictors.span_qa_predictors import APISpanQAPredictor
from papermage.predictors.token_predictors import HFWhitspaceTokenPredictor
from papermage.predictors.vila_predictors import IVILATokenClassificationPredictor
from papermage.predictors.word_predictors import SVMWordPredictor

__all__ = [
    "HFBIOTaggerPredictor",
    "IVILATokenClassificationPredictor",
    "HFWhitspaceTokenPredictor",
    "SVMWordPredictor",
    "PysbdSentencePredictor",
    "LPEffDetPubLayNetBlockPredictor",
    "LPEffDetFormulaPredictor",
    "APISpanQAPredictor",
    "BasePredictor",
]



"""

Create a Document using only text <str> input. Useful for test and development.

@kylel

"""

import logging
from pathlib import Path
from typing import Dict, List, Union

from papermage.magelib import Box, Document, Entity, SentencesFieldName, TokensFieldName
from papermage.predictors import HFWhitspaceTokenPredictor, PysbdSentencePredictor
from papermage.recipes.recipe import Recipe


class TextRecipe(Recipe):
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.tokenizer_predictor = HFWhitspaceTokenPredictor()
        self.sent_predictor = PysbdSentencePredictor()

    def from_str(self, text: str) -> Document:
        doc = Document(symbols=text)
        return self.from_doc(doc=doc)

    def from_doc(self, doc: Document) -> Document:
        self.logger.info("Predicting tokens...")
        tokens = self.tokenizer_predictor.predict(doc=doc)
        doc.annotate_layer(name=TokensFieldName, entities=tokens)

        self.logger.info("Predicting sentences...")
        sentences = self.sent_predictor.predict(doc=doc)
        doc.annotate_layer(name=SentencesFieldName, entities=sentences)
        return doc



"""

A (start, end) interval that references some string. For example, span (0, 4)
represents the word 'This' in the string 'This is a document.'

@kylel, @egork

"""

from collections import defaultdict
from typing import Dict, List


class Span:
    __slots__ = ["start", "end"]

    def __init__(self, start: int, end: int):
        self.start = start
        self.end = end

    def to_json(self) -> List[int]:
        """Returns whatever representation is JSON compatible"""
        return [self.start, self.end]

    @classmethod
    def from_json(cls, span_json: List) -> "Span":
        """Recreates the object from the JSON serialization"""
        return Span(start=span_json[0], end=span_json[-1])

    def __repr__(self):
        return f"Span{self.to_json()}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Span):
            return False
        return self.start == other.start and self.end == other.end

    def __lt__(self, other: "Span"):
        """Useful for sort(). Orders according to the start index.
        If ties, then order according to the end index."""
        if self.start == other.start:
            return self.end < other.end
        return self.start < other.start

    def __hash__(self) -> int:
        return hash((self.start, self.end))

    def is_overlap(self, other: "Span") -> bool:
        """Whether self overlaps with the other Span object."""
        return self.start <= other.start < self.end or other.start <= self.start < other.end or self == other

    @classmethod
    def create_enclosing_span(cls, spans: List["Span"]) -> "Span":
        """Create the narrowest Span that completely encloses all the input Spans."""
        if not spans:
            raise ValueError(f"`spans` should be non-empty.")
        start = spans[0].start
        end = spans[0].end
        for span in spans[1:]:
            if span.start < start:
                start = span.start
            if span.end > end:
                end = span.end
        return Span(start=start, end=end)




from papermage.recipes import CoreRecipe

recipe = CoreRecipe()
doc = recipe.run("/Users/kylel/ai2/paperMage/tests/fixtures/papermage.pdf")

# problematic section
weird_section = doc.sections[1]
print(weird_section)

# it's split between 3 rows
for row in weird_section.rows:
    print(row)

# but visually, this should all be one big row. why is it split?
from papermage.visualizers.visualizer import plot_entities_on_page

page_with_weird_section = weird_section.pages[0].id
plot_entities_on_page(page_image=doc.images[page_with_weird_section], entities=weird_section.rows)




"""

@kylel, benjaminn

"""

import json
import pathlib
import unittest

import transformers

from papermage.magelib import Document, Entity, Span
from papermage.predictors.base_predictors.hf_predictors import (
    BIOBatch,
    BIOPrediction,
    HFBIOTaggerPredictor,
)

TEST_SCIBERT_WEIGHTS = "allenai/scibert_scivocab_uncased"


class TestBioTaggerPredictor(unittest.TestCase):
    def setUp(self):
        transformers.set_seed(407)
        self.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"

        # setup document
        with open(self.fixture_path / "entity_classification_predictor_test_doc_papermage.json", "r") as f:
            test_doc_json = json.load(f)
        self.doc = Document.from_json(doc_json=test_doc_json)
        ent1 = Entity(spans=[Span(start=86, end=456)])
        ent2 = Entity(spans=[Span(start=457, end=641)])
        self.doc.annotate_layer(name="bibs", entities=[ent1, ent2])

        # setup predictor
        self.id2label = {0: "O", 1: "B_Label", 2: "I_Label"}
        self.label2id = {label: id_ for id_, label in self.id2label.items()}
        self.predictor = HFBIOTaggerPredictor.from_pretrained(
            model_name_or_path=TEST_SCIBERT_WEIGHTS,
            entity_name="tokens",
            context_name="pages",
            **{"num_labels": len(self.id2label), "id2label": self.id2label, "label2id": self.label2id},
        )

    def test_preprocess(self):
        doc = Document(symbols="This is a test document.")
        tokens = [
            Entity(spans=[Span(start=0, end=4)]),
            Entity(spans=[Span(start=5, end=7)]),
            Entity(spans=[Span(start=8, end=9)]),
            Entity(spans=[Span(start=10, end=14)]),
            Entity(spans=[Span(start=15, end=23)]),
            Entity(spans=[Span(start=23, end=24)]),
        ]
        doc.annotate_layer(name="tokens", entities=tokens)
        sents = [Entity(spans=[Span(start=0, end=24)])]
        doc.annotate_layer(name="sents", entities=sents)

        batches = self.predictor.preprocess(doc=doc, context_name="sents")
        self.assertIsInstance(batches[0], BIOBatch)
        decoded_batch = self.predictor.tokenizer.batch_decode(batches[0].input_ids)
        self.assertListEqual(decoded_batch, ["[CLS] this is a test document. [SEP]"])

    def test_predict_pages_tokens(self):
        predictor = HFBIOTaggerPredictor.from_pretrained(
            model_name_or_path=TEST_SCIBERT_WEIGHTS,
            entity_name="tokens",
            context_name="pages",
            **{"num_labels": len(self.id2label), "id2label": self.id2label, "label2id": self.label2id},
        )
        token_tags = predictor.predict(doc=self.doc)
        assert len(token_tags) == 340

        self.doc.annotate_layer(name="token_tags", entities=token_tags)
        for token_tag in token_tags:
            assert isinstance(token_tag.metadata.label, str)
            assert isinstance(token_tag.metadata.score, float)

    def test_predict_bibs_tokens(self):
        self.predictor.context_name = "bibs"
        token_tags = self.predictor.predict(doc=self.doc)
        assert len(token_tags) == 38

    def test_missing_fields(self):
        self.predictor.entity_name = "OHNO"
        with self.assertRaises(AssertionError) as e:
            self.predictor.predict(doc=self.doc)
            assert "OHNO" in e.exception

        self.predictor.entity_name = "tokens"
        self.predictor.context_name = "BLABLA"
        with self.assertRaises(AssertionError) as e:
            self.predictor.predict(doc=self.doc)
            assert "BLABLA" in e.exception

        self.predictor.context_name = "pages"

    def test_predict_pages_tokens_roberta(self):
        predictor = HFBIOTaggerPredictor.from_pretrained(
            model_name_or_path="roberta-base",
            entity_name="tokens",
            context_name="pages",
            add_prefix_space=True,  # Needed for roberta
            **{"num_labels": len(self.id2label), "id2label": self.id2label, "label2id": self.label2id},
        )
        token_tags = predictor.predict(doc=self.doc)
        assert len(token_tags) == 924

        self.doc.annotate_layer(name="token_tags", entities=token_tags)
        for token_tag in token_tags:
            assert isinstance(token_tag.metadata.label, str)
            assert isinstance(token_tag.metadata.score, float)

    # def test_postprocess(self):
    #     self.predictor.postprocess(
    #         doc=self.doc,
    #         context_name="pages",
    #         preds=[
    #             BIOPrediction(context_id=0, entity_id=0, label="B-Label", score=0.4),
    #             BIOPrediction(context_id=0, entity_id=1, label="I-Label", score=0.2),
    #             BIOPrediction(context_id=0, entity_id=2, label="O", score=0.3),
    #             BIOPrediction(context_id=0, entity_id=3, label=None, score=None),
    #             BIOPrediction(context_id=0, entity_id=4, label="B-Label", score=0.4),
    #             BIOPrediction(context_id=0, entity_id=5, label=None, score=None),
    #             BIOPrediction(context_id=0, entity_id=6, label="I-Label", score=0.2),
    #         ],
    #     )



"""

Formula predictors.

@kylel

"""

import pathlib
import unittest

from papermage.predictors import LPEffDetFormulaPredictor


class TestLayoutParserBoxMathFormulaBlockPredictor(unittest.TestCase):
    def setUp(self):
        self.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"
        self.predictor = LPEffDetFormulaPredictor.from_pretrained()

    def test_predict(self):
        pass



"""


@kylel

"""

import logging
from itertools import chain
from typing import Dict, List, NamedTuple, Optional, Tuple, Union

from .box import Box
from .entity import Entity
from .image import Image
from .layer import Layer
from .metadata import Metadata
from .names import (
    EntitiesFieldName,
    ImagesFieldName,
    MetadataFieldName,
    RelationsFieldName,
    SymbolsFieldName,
    TokensFieldName,
)
from .span import Span


class Prediction(NamedTuple):
    name: str
    entities: List[Entity]


class Document:
    tokens: Layer
    rows: Layer
    blocks: Layer
    words: Layer
    sentences: Layer
    paragraphs: Layer
    pages: Layer

    SPECIAL_FIELDS = [SymbolsFieldName, ImagesFieldName, MetadataFieldName]

    def __init__(
        self,
        symbols: Optional[str] = None,
        images: Optional[List[Image]] = None,
        metadata: Optional[Metadata] = None,
    ):
        self.symbols = symbols if symbols else None
        self.images = images if images else None
        if not self.symbols and not self.images:
            raise ValueError("Document must have at least one of `symbols` or `images`")
        self.metadata = metadata if metadata else Metadata()
        self._layers: List[str] = []

    @property
    def layers(self) -> List[str]:
        return self.SPECIAL_FIELDS + self._layers

    def validate_layer_name_availability(self, name: str) -> None:
        if name in self.layers:
            raise AssertionError(f"{name} not allowed Document.SPECIAL_FIELDS.")
        if name in self.layers:
            raise AssertionError(f'{name} already exists. Try `doc.remove_layer("{name}")` first.')
        if name in dir(self):
            raise AssertionError(f"{name} clashes with Document class properties.")

    def get_layer(self, name: str) -> Layer:
        """Gets a layer by name. For example, `doc.get_layer("sentences")` returns sentences."""
        return getattr(self, name)

    def annotate(self, *predictions: Union[Prediction, Tuple[Prediction, ...]]) -> None:
        """Annotates the document with predictions."""
        all_preds = chain.from_iterable([p] if isinstance(p, Prediction) else p for p in predictions)
        for prediction in all_preds:
            self.annotate_layer(name=prediction.name, entities=prediction.entities)

    def annotate_layer(self, name: str, entities: Union[List[Entity], Layer]) -> None:
        self.validate_layer_name_availability(name=name)

        if isinstance(entities, list):
            layer = Layer(entities=entities)
        else:
            layer = entities

        layer.doc = self
        layer.name = name
        setattr(self, name, layer)
        self._layers.append(name)

    def remove_layer(self, name: str):
        if name not in self.layers:
            pass
        else:
            getattr(self, name).doc = None
            getattr(self, name).name = None
            delattr(self, name)
            self._layers.remove(name)

    def get_relation(self, name: str) -> List["Relation"]:
        raise NotImplementedError

    def annotate_relation(self, name: str) -> None:
        self.validate_layer_name_availability(name=name)
        raise NotImplementedError

    def remove_relation(self, name: str) -> None:
        raise NotImplementedError

    def annotate_images(self, images: List[Image]) -> None:
        if len(images) == 0:
            raise ValueError("No images were provided")

        image_types = {type(image) for image in images}
        if len(image_types) > 1:
            raise TypeError(f"Images contain multiple types: {image_types}")
        image_type = image_types.pop()

        if not issubclass(image_type, Image):
            raise NotImplementedError(f"Unsupported image type {image_type} for {ImagesFieldName}")

        setattr(self, ImagesFieldName, images)

    def remove_images(self) -> None:
        raise NotImplementedError

    def to_json(self, layers: Optional[List[str]] = None, with_images: bool = False) -> Dict:
        """Returns a dictionary that's suitable for serialization

        Use `fields` to specify a subset of groups in the Document to include (e.g. 'sentences')

        Output format looks like
            {
                symbols: "...",
                entities: {...},
                relations: {...},
                metadata: {...}
            }
        """
        # 1) instantiate basic Document dict
        doc_dict = {
            SymbolsFieldName: self.symbols,
            MetadataFieldName: self.metadata.to_json(),
            EntitiesFieldName: {},
            RelationsFieldName: {},
        }

        # 2) serialize each layer to JSON
        layers = self._layers if layers is None else layers
        for layer in layers:
            doc_dict[EntitiesFieldName][layer] = [entity.to_json() for entity in getattr(self, layer)]

        # 3) serialize images if `with_images == True`
        if with_images:
            doc_dict[ImagesFieldName] = [image.to_base64() for image in getattr(self, ImagesFieldName)]

        return doc_dict

    @classmethod
    def from_json(cls, doc_json: Dict) -> "Document":
        # 1) instantiate basic Document
        symbols = doc_json[SymbolsFieldName]
        doc = cls(symbols=symbols, metadata=Metadata(**doc_json.get(MetadataFieldName, {})))

        # 2) instantiate entities
        for field_name, entity_jsons in doc_json[EntitiesFieldName].items():
            entities = [Entity.from_json(entity_json=entity_json) for entity_json in entity_jsons]
            doc.annotate_layer(name=field_name, entities=entities)

        return doc

    def __repr__(self):
        return f"Document with {len(self.layers)} layers: {self.layers}"

    def find(self, query: Union[Span, Box], name: str) -> List[Entity]:
        """Finds all entities that intersect with the query"""
        logger = logging.getLogger(__name__)
        logger.warning(
            "This method is deprecated due to ambiguity and will be removed in a future release."
            "Please use Document.intersect_by_span or Document.intersect_by_box instead."
        )
        if isinstance(query, Span):
            return self.intersect_by_span(query=Entity(spans=[query]), name=name)
        elif isinstance(query, Box):
            return self.intersect_by_box(query=Entity(boxes=[query]), name=name)
        else:
            raise TypeError(f"Unsupported query type {type(query)}")

    def intersect_by_span(self, query: Entity, name: str) -> List[Entity]:
        """Finds all entities that intersect with the query"""
        return self.get_layer(name=name).intersect_by_span(query=query)

    def intersect_by_box(self, query: Entity, name: str) -> List[Entity]:
        """Finds all entities that intersect with the query"""
        return self.get_layer(name=name).intersect_by_box(query=query)



"""

Layers are collections of Entities. Supports indexing and slicing.

@kylel

"""

import logging
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from .box import Box
from .entity import Entity
from .indexer import EntityBoxIndexer, EntitySpanIndexer
from .metadata import Metadata
from .span import Span

if TYPE_CHECKING:
    from .document import Document


class Layer:
    """A fancy list of Entities. Manages <list> things like indexing and slicing,
    but also gives access to things like reading order and other metadata."""

    __slots__ = ["entities", "metadata", "span_indexer", "box_indexer", "_name", "_doc"]

    def __init__(self, entities: List[Entity], metadata: Optional[Metadata] = None):
        self.entities = entities
        for i, entity in enumerate(entities):
            entity.layer = self
            entity.id = i
        self.metadata = metadata if metadata else Metadata()
        self.span_indexer = EntitySpanIndexer(entities=entities)
        self.box_indexer = EntityBoxIndexer(entities=entities)
        self._name = None
        self._doc = None

    def __repr__(self):
        entity_repr = "\n".join([f"\t{e}" for e in self.entities])
        return f"Layer with {len(self)} Entities:\n{entity_repr}"

    def __getitem__(self, key):
        return self.entities[key]

    def __len__(self):
        return len(self.entities)

    def __iter__(self):
        return iter(self.entities)

    def __contains__(self, item):
        return item in self.entities

    def __eq__(self, other: "Layer") -> bool:
        """Layers are equal if all elements are equal"""
        return self.entities == other.entities

    def to_json(self):
        return [entity.to_json() for entity in self.entities]

    @classmethod
    def from_json(cls, layer_json):
        return cls(entities=[Entity.from_json(entity_json) for entity_json in layer_json])

    def find(self, query: Union[Span, Box]) -> List[Entity]:
        logger = logging.getLogger(__name__)
        logger.warning(
            "This method is deprecated due to ambiguity and will be removed in a future release."
            "Please use Layer.intersect_by_span or Layer.intersect_by_box instead."
        )
        if isinstance(query, Span):
            return self.intersect_by_span(query=Entity(spans=[query]))
        elif isinstance(query, Box):
            return self.intersect_by_box(query=Entity(boxes=[query]))
        else:
            raise TypeError(f"Unsupported query type {type(query)}")

    def intersect_by_span(self, query: Entity) -> List[Entity]:
        return self.span_indexer.find(query=query)

    def intersect_by_box(self, query: Entity) -> List[Entity]:
        return self.box_indexer.find(query=query)

    @property
    def doc(self) -> Optional["Document"]:
        return self._doc

    @doc.setter
    def doc(self, doc: Optional["Document"]) -> None:
        """This method attaches a Document to this Layer, allowing the Layer
        to access things beyond itself within the Document (e.g. other layers)"""
        if self.doc and doc:
            raise AttributeError(
                "Already has an attached Document. Since Layer should correspond"
                "to only a specific Document, we recommend creating a new"
                "Layer from scratch and then attaching your Document."
            )
        self._doc = doc

    @property
    def name(self) -> Optional[int]:
        return self._name

    @name.setter
    def name(self, name: Optional[str]) -> None:
        if self.name and name:
            raise AttributeError(f"This Layer already has a name: {self.name}")
        self._name = name

    def __getattr__(self, name: str) -> List["Entity"]:
        if not self.doc:
            raise AttributeError("This Layer is missing a Document")

        if name not in self.doc.layers:
            raise AttributeError(f"Layer {name} not found in Document")

        return self.crossref_layers_by_span(source=self, target=self.doc.get_layer(name=name))

    @classmethod
    def crossref_layers_by_span(cls, source: "Layer", target: "Layer") -> List[Entity]:
        """This method allows you to access overlapping Entities between two Layers"""
        raise NotImplementedError



"""

@kylel

"""

import pathlib
import unittest

from papermage.recipes import TextRecipe


class TestCoreRecipe(unittest.TestCase):
    def setUp(self):
        self.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"
        self.recipe = TextRecipe()

    def test_stability(self):
        doc = self.recipe.run("This is a test document. This is a test document. This is a test document.")
        self.assertEqual(doc.symbols, "This is a test document. This is a test document. This is a test document.")
        self.assertEqual(len(doc.tokens), 15)
        self.assertEqual(len(doc.sentences), 3)
        self.assertEqual(doc.tokens[0].text, "This")
        self.assertEqual(doc.sentences[0].text, "This is a test document.")



"""

An annotated "unit" in a Layer.

"""

import logging
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from .box import Box
from .image import Image
from .metadata import Metadata
from .names import TokensFieldName
from .span import Span

if TYPE_CHECKING:
    from .layer import Layer


class Entity:
    __slots__ = ["spans", "boxes", "images", "metadata", "_id", "_layer"]

    def __init__(
        self,
        spans: Optional[List[Span]] = None,
        boxes: Optional[List[Box]] = None,
        images: Optional[List[Image]] = None,
        metadata: Optional[Metadata] = None,
    ):
        if not spans and not boxes:
            raise ValueError(f"At least one of `spans` or `boxes` must be set.")
        self.spans = spans if spans is not None else []
        self.boxes = boxes if boxes is not None else []
        self.images = images if images is not None else []
        self.metadata = metadata if metadata else Metadata()
        # TODO: it's confusing that `id` is both reading order as well as direct reference
        # TODO: maybe Layer() should house reading order, and Entity() should have a unique ID
        # TODO: hashing would be interesting, but Metadata() is allowed to mutate so that's a problem
        self._id = None
        self._layer = None

    def __repr__(self):
        if self.layer:
            return f"Annotated Entity:\tID: {self.id}\tSpans: {True if self.spans else False}\tBoxes: {True if self.boxes else False}\tText: {self.text}"
        return f"Unannotated Entity: {self.to_json()}"

    def to_json(self) -> Dict:
        entity_dict = dict(
            spans=[span.to_json() for span in self.spans],
            boxes=[box.to_json() for box in self.boxes],
            metadata=self.metadata.to_json(),
        )
        # only serialize non-null/non-empty values
        return {k: v for k, v in entity_dict.items() if v}

    @classmethod
    def from_json(cls, entity_json: Dict) -> "Entity":
        return cls(
            spans=[Span.from_json(span_json=span_json) for span_json in entity_json.get("spans", [])],
            boxes=[Box.from_json(box_json=box_json) for box_json in entity_json.get("boxes", [])],
            metadata=Metadata.from_json(entity_json.get("metadata", {})),
        )

    @property
    def layer(self) -> Optional["Layer"]:
        return self._layer

    @layer.setter
    def layer(self, layer: Optional["Layer"]) -> None:
        """This method attaches a Layer to this Entity, allowing the Entity
        to access things beyond itself within the Layer (e.g. neighboring Entities)"""
        if self.layer and layer:
            raise AttributeError(
                "Already has an attached Layer. Since Entity should correspond"
                "to only a specific Layer, we recommend creating a new"
                "Entity from scratch and then attaching your Layer."
            )
        self._layer = layer

    @property
    def id(self) -> Optional[int]:
        return self._id

    @id.setter
    def id(self, id: int) -> None:
        """This method assigns an ID to an Entity. Requires a Document to be attached
        to this Entity. ID basically gives the Entity itself awareness of its
        position within the broader Document."""
        if self.id:
            raise AttributeError(f"This Entity already has an ID: {self.id}")
        if not self.layer:
            raise AttributeError("This Entity is missing a Document")
        self._id = id

    def __getattr__(self, name: str) -> List["Entity"]:
        """This Overloading is convenient syntax since the `entity.layer` operation is intuitive for folks."""
        # add method deprecation warning
        logger = logging.getLogger(__name__)
        logger.warning(
            "Entity.__getattr__ is deprecated due to ambiguity and will be removed in a future release."
            "Please use Entity.intersect_by_span or Entity.intersect_by_box instead."
        )
        try:
            if len(self.spans) > 0:
                intersection = self.intersect_by_span(name=name)
                if len(intersection) == 0 and len(self.boxes) > 0:
                    intersection = self.intersect_by_box(name=name)
                return intersection
            else:
                return self.intersect_by_box(name=name)
        except ValueError:
            # maybe users just want some attribute of the Entity object
            return self.__getattribute__(name)

    def intersect_by_span(self, name: str) -> List["Entity"]:
        """This method allows you to access overlapping Entities
        within the Document based on Span"""
        if self.layer is None:
            raise ValueError("This Entity is not attached to a Layer")

        if self.layer.doc is None:
            raise ValueError("This Entity's Layer is not attached to a Document")

        return self.layer.doc.intersect_by_span(query=self, name=name)

    def intersect_by_box(self, name: str) -> List["Entity"]:
        """This method allows you to access overlapping Entities
        within the Document based on Box"""
        if self.layer is None:
            raise ValueError("This Entity is not attached to a Layer")

        if self.layer.doc is None:
            raise ValueError("This Entity's Layer is not attached to a Document")

        return self.layer.doc.intersect_by_box(query=self, name=name)

    @property
    def start(self) -> Union[int, float]:
        return min([span.start for span in self.spans]) if len(self.spans) > 0 else float("-inf")

    @property
    def end(self) -> Union[int, float]:
        return max([span.end for span in self.spans]) if len(self.spans) > 0 else float("inf")

    @property
    def symbols_from_spans(self) -> List[str]:
        if self.layer is None:
            raise ValueError("This Entity is not attached to a Layer")

        if self.layer.doc is None:
            raise ValueError("This Entity's Layer is not attached to a Document")

        if self.layer.doc.symbols is None:
            raise ValueError("This Entity's Document is missing symbols")

        return [self.layer.doc.symbols[span.start : span.end] for span in self.spans]

    @property
    def symbols_from_boxes(self) -> List[str]:
        if self.layer is None:
            raise ValueError("This Entity is not attached to a Layer")

        if self.layer.doc is None:
            raise ValueError("This Entity's Layer is not attached to a Document")

        if self.layer.doc.symbols is None:
            raise ValueError("This Entity's Document is missing symbols")

        matched_tokens = self.intersect_by_box(name=TokensFieldName)
        return [self.layer.doc.symbols[span.start : span.end] for t in matched_tokens for span in t.spans]

    @property
    def text(self) -> str:
        # return stored metadata
        maybe_text = self.metadata.get("text", None)
        if maybe_text:
            return maybe_text
        # return derived from symbols
        if self.symbols_from_spans:
            return " ".join(self.symbols_from_spans).replace("\n", " ")
        # return derived from boxes and tokens
        if self.symbols_from_boxes:
            return " ".join(self.symbols_from_boxes).replace("\n", " ")
        return ""

    @text.setter
    def text(self, text: Union[str, None]) -> None:
        self.metadata.text = text

    def __iter__(self):
        """By default, iterate over the spans"""
        yield from self.spans

    def __lt__(self, other: "Entity"):
        if self.id and other.id:
            return self.id < other.id
        else:
            return self.start < other.start



"""

Needs to be imported in this order.

@kylel

"""

from .box import Box
from .document import Document, Prediction
from .entity import Entity
from .image import Image
from .indexer import EntityBoxIndexer, EntitySpanIndexer
from .layer import Layer
from .metadata import Metadata
from .names import (
    AbstractsFieldName,
    AlgorithmsFieldName,
    AuthorsFieldName,
    BibliographiesFieldName,
    BlocksFieldName,
    CaptionsFieldName,
    EntitiesFieldName,
    EquationsFieldName,
    FiguresFieldName,
    FootersFieldName,
    FootnotesFieldName,
    HeadersFieldName,
    ImagesFieldName,
    KeywordsFieldName,
    ListsFieldName,
    MetadataFieldName,
    PagesFieldName,
    ParagraphsFieldName,
    RelationsFieldName,
    RowsFieldName,
    SectionsFieldName,
    SentencesFieldName,
    SymbolsFieldName,
    TablesFieldName,
    TitlesFieldName,
    TokensFieldName,
    WordsFieldName,
)
from .span import Span

__all__ = [
    "AbstractsFieldName",
    "AlgorithmsFieldName",
    "AuthorsFieldName",
    "BibliographiesFieldName",
    "BlocksFieldName",
    "Box",
    "Prediction",
    "CaptionsFieldName",
    "Document",
    "EntitiesFieldName",
    "Entity",
    "EntityBoxIndexer",
    "EntitySpanIndexer",
    "EquationsFieldName",
    "FiguresFieldName",
    "FootersFieldName",
    "FootnotesFieldName",
    "HeadersFieldName",
    "Image",
    "ImagesFieldName",
    "KeywordsFieldName",
    "KeywordsFieldName",
    "Layer",
    "ListsFieldName",
    "Metadata",
    "MetadataFieldName",
    "PagesFieldName",
    "ParagraphsFieldName",
    "RelationsFieldName",
    "RowsFieldName",
    "SectionsFieldName",
    "SentencesFieldName",
    "Span",
    "SymbolsFieldName",
    "TablesFieldName",
    "TitlesFieldName",
    "TokensFieldName",
    "WordsFieldName",
]
